from __future__ import annotations

import re
from http import HTTPStatus

from .exceptions import ParserInvalidDataError
from .utils import ensure_length

SENTINEL = object()


class HeadersMultiValue(list):
    pass


class Headers(dict):
    def __setitem__(self, key, value):
        key = key.lower()
        v = self.get(key, SENTINEL)
        if v is SENTINEL:
            return super().__setitem__(key, value)
        elif type(v) is HeadersMultiValue:
            return v.append(value)
        else:
            nv = HeadersMultiValue()
            nv.append(v)
            nv.append(value)
            return super().__setitem__(key, nv)

    def __getitem__(self, key, value):
        return super().__getitem__(key.lower(), value)

    def __delitem__(self, key, value):
        return super().__delitem__(key.lower(), value)

    def get(self, key, default=None):
        return super().get(key.lower(), default)

    def getone(self, key, default=None):
        value = self.get(key, default)
        if type(value) is HeadersMultiValue:
            value = value[0]
        return value

    def pop(self, key, *args, **kwargs):
        return super().pop(key.lower(), *args, **kwargs)

    def popone(self, key, default=SENTINEL):
        value = self.get(key, default)
        if value is SENTINEL:
            raise KeyError(key)
        elif type(value) is HeadersMultiValue:
            value = value.pop(0)
        else:
            del self[key]
        return value


class HTTPResponseProtocol:
    def http_response_received(self, response: HTTPResponse) -> None:
        pass


class HTTPRequestProtocol:
    def http_request_received(self, request: HTTPRequest) -> None:
        pass


def _iter_headers(headers: bytes):
    offset = 0

    while True:
        index = headers.index(b'\r\n', offset) + 2
        data = headers[offset:index]
        offset = index

        if data == b'\r\n':
            return

        yield [item.strip() for item in data.split(b':', 1)]


def _find_headers():
    data = b''

    while True:
        data += yield
        end = data.find(b'\r\n\r\n') + 4

        if end != -1:
            headers = data[:end]
            body = data[end:]
            return _iter_headers(headers), body


class HTTPResponse:
    STATUS_LINE_REGEX = re.compile(
        r'HTTP/(?P<version>\d((?=\.)(\.\d))?) (?P<status>\d+) (?P<phrase>.+)')

    def __init__(self, *, version='1.1', status, phrase, headers, body):
        self.version = version
        self.status = HTTPStatus(status)
        self.phrase = phrase
        self.headers = headers
        self.body = body

    def encode(self) -> bytes:
        response = [f'HTTP/{self.version} {self.status} {self.status.phrase}']
        response.extend(f'{k}: {v}' for k, v in self.headers.items())
        response.append('\r\n')
        return b'\r\n'.join(part.encode() for part in response)

    @classmethod
    def parser(cls, protocol: HTTPResponseProtocol):
        headers, body = yield from _find_headers()

        status_line, = next(headers)

        extra = {
            'protocol': protocol,
            'status_line': status_line
        }

        try:
            match = cls.STATUS_LINE_REGEX.match(status_line.decode())
        except UnicodeDecodeError as e:
            raise ParserInvalidDataError(str(e), extra)

        if match is None:
            raise ParserInvalidDataError('The status line is invalid', extra)

        headers_dict = Headers()
        for key, value in headers:
            headers_dict[key] = value

        content_length = headers_dict.get(b'content-length')

        if content_length is not None:
            try:
                content_length = int(content_length.decode())
            except UnicodeDecodeError as e:
                raise ParserInvalidDataError(str(e), extra)
        else:
            content_length = 0

        body = yield from ensure_length(body, content_length)

        response = cls(version=match.group('version'),
                       status=int(match.group('status')),
                       phrase=match.group('phrase'),
                       headers=headers_dict,
                       body=body[:content_length])
        protocol.http_response_received(response)

        return body[content_length:]


class HTTPRequest:
    METHODS = ('GET', 'HEAD', 'POST', 'PUT', 'DELETE',
               'CONNECT', 'OPTIONS', 'TRACE', 'PATCH')

    REQUEST_LINE_REXEX = re.compile(
        rf'(?P<method>{"|".join(METHODS)}) (?P<path>.+) '
        r'HTTP/(?P<version>\d((?=\.)(\.\d))?)')

    def __init__(self, *, version='1.1', method, path=None, headers, body):
        self.version = version
        self.method = method
        self.path = path or '/'
        self.headers = headers
        self.body = body

    def encode(self) -> bytes:
        request = [f'{self.method} {self.path} HTTP/{self.version}']
        request.extend(f'{k}: {v}' for k, v in self.headers.items())
        request.append('\r\n')
        return b'\r\n'.join(part.encode() for part in request)

    @classmethod
    def parser(cls, protocol: HTTPRequestProtocol):
        headers, body = yield from _find_headers()

        request_line, = next(headers)

        extra = {
            'protocol': protocol,
            'request_line': request_line
        }

        try:
            match = cls.REQUEST_LINE_REXEX.match(request_line.decode())
        except UnicodeDecodeError as e:
            raise ParserInvalidDataError(str(e), extra)

        if match is None:
            raise ParserInvalidDataError('The request line is invalid', extra)

        headers_dict = Headers()
        for key, value in headers:
            headers_dict[key] = value

        content_length = headers_dict.get(b'content-length', 0)

        body = yield from ensure_length(body, content_length)

        request = cls(version=match.group('version'),
                      method=match.group('method'),
                      path=match.group('path'),
                      headers=headers_dict,
                      body=body)
        protocol.http_request_received(request)

        return body[content_length:]



import base64
import os
import urllib.parse
from http import HTTPStatus

from .exceptions import BrokenHandshakeError, WsaioError
from .http import HTTPRequest, HTTPResponse, HTTPResponseProtocol
from .protocol import BaseProtocol, BaseProtocolState, taskify
from .websocket import (WebSocketCloseCode, WebSocketFrame, WebSocketOpcode,
                        WebSocketProtocol, WebSocketState)


class WebSocketClient(BaseProtocol, HTTPResponseProtocol, WebSocketProtocol):
    def __init__(self, loop=None):
        BaseProtocol.__init__(self, loop)
        HTTPResponseProtocol.__init__(self)
        WebSocketProtocol.__init__(self)
        self._handshake_complete = self.loop.create_future()

    strstate = WebSocketProtocol.strstate

    def http_response_received(self, response: HTTPResponse) -> None:
        extra = {
            'response': response,
            'protocol': self
        }

        expected_status = HTTPStatus.SWITCHING_PROTOCOLS
        if response.status is not expected_status:
            return self._close(
                BrokenHandshakeError(
                    f'Server responsed with status code {response.status} '
                    f'({response.phrase}), need status code {expected_status} '
                    f'({expected_status.phrase}) to complete handshake. '
                    'Closing!',
                    extra
                )
            )

        connection = response.headers.getone(b'connection')
        if connection is None or connection.lower() != b'upgrade':
            return self._close(
                BrokenHandshakeError(
                    f'Server responded with "connection: {connection}", '
                    f'need "connection: upgrade" to complete handshake. '
                    'Closing!',
                    extra
                )
            )

        upgrade = response.headers.getone(b'upgrade')
        if upgrade is None or upgrade.lower() != b'websocket':
            return self._close(
                BrokenHandshakeError(
                    f'Server responded with "upgrade: {upgrade}", '
                    f'need "upgrade: websocket" to complete handshake. '
                    'Closing!',
                    extra
                )
            )

        self.state = BaseProtocolState.IDLE

        self.set_parser(WebSocketFrame.parser(self))
        self._handshake_complete.set_result(None)

        self.ws_connected()

    @taskify
    async def connection_made(self, transport):
        super().connection_made(transport)
        request = HTTPRequest(
            method='GET',
            path=self.url.path + self.url.params,
            headers=self.headers,
            body=b''
        )
        await self.write(request.encode())

    async def connect(self, url, *args, **kwargs):
        self.sec_ws_key = base64.b64encode(os.urandom(16))

        self.headers = kwargs.pop('headers', {})

        self.set_parser(HTTPResponse.parser(self))

        self.url = urllib.parse.urlparse(url)
        self.ssl = kwargs.pop('ssl', self.url.scheme == 'wss')
        self.port = kwargs.pop('port', 443 if self.ssl else 80)

        self.headers.update({
            'Host': f'{self.url.hostname}:{self.port}',
            'Connection': 'Upgrade',
            'Upgrade': 'websocket',
            'Sec-WebSocket-Key': self.sec_ws_key.decode(),
            'Sec-WebSocket-Version': 13
        })

        self.state = WebSocketState.HANDSHAKING

        await self.loop.create_connection(
            lambda: self, self.url.hostname,
            self.port, *args, ssl=self.ssl, **kwargs
        )

        await self._handshake_complete

    def parser_invalid_data(self, exc):
        self._close(exc)

    @taskify
    async def _close(self, exc=None):
        if exc is not None:
            if self.state is WebSocketState.HANDSHAKING:
                self._handshake_complete.set_exception(exc)
            else:
                close_code = WebSocketCloseCode.NORMAL_CLOSURE
                if isinstance(exc, WsaioError):
                    close_code = exc.get_extra(
                        'close_code',
                        WebSocketCloseCode.NORMAL_CLOSURE
                    )
                await self._send_close(close_code, str(exc).encode())
        super()._close(exc)

    async def _send_close(self, code: int, data: bytes, *,
                          drain: bool = True) -> None:
        code = code.to_bytes(2, 'big', signed=False)
        await self.send_frame(
            WebSocketFrame(opcode=WebSocketOpcode.CLOSE,
                           data=code + (data or b'')),
            drain=drain)

    async def close(self, code: int, data: bytes = None, *,
                    drain: bool = True) -> None:
        await self._send_close(code, data, drain=drain)
        super()._close()

    async def send_frame(self, frame: WebSocketFrame, **kwargs) -> None:
        await self.write(frame.encode(masked=True), **kwargs)

    async def send_bytes(self, data: bytes, *,
                         opcode: WebSocketOpcode = WebSocketOpcode.TEXT,
                         **kwargs) -> None:
        await self.send_frame(WebSocketFrame(opcode=opcode, data=data),
                              **kwargs)

    async def send_str(self, data: str, *args, **kwargs) -> None:
        await self.send_bytes(data.encode(), *args, **kwargs)



import asyncio
import enum
import functools

from .exceptions import ConnectionClosedError, ParserInvalidDataError


class BaseProtocolState(enum.IntEnum):
    INIT = 0
    IDLE = 1
    CLOSED = 3
    PARSING = 2


class BaseProtocol(asyncio.Protocol):
    # Stolen from asyncio.streams.FlowControlMixin
    def __init__(self, loop=None):
        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop
        self._paused = False
        self._drain_waiter = None
        self._connection_lost = False
        self._parser = None
        self.transport = None
        self.state = BaseProtocolState.INIT

    def strstate(self):
        if self.state is BaseProtocolState.INIT:
            return 'initialized'
        elif self.state is BaseProtocolState.IDLE:
            return 'idling'
        elif self.state is BaseProtocolState.PARSING:
            return 'parsing'
        elif self.state is BaseProtocolState.CLOSED:
            return 'closed'

    def _close(self, exc=None):
        self.closing_connection(exc)
        self.state = BaseProtocolState.CLOSED
        self.transport.close()

    def connection_made(self, transport):
        self.transport = transport

    def set_parser(self, parser):
        parser.send(None)
        self._parser = parser

    def data_received(self, data):
        state = self.state
        self.state = BaseProtocolState.PARSING
        while data:
            try:
                self._parser.send(data)
                self.state = state
                return
            except StopIteration as e:
                # the parser must have changed, e.value is the unused data
                data = e.value
            except ParserInvalidDataError as e:
                self._close(e)
                return
        self.state = state

    def pause_writing(self):
        assert not self._paused
        self._paused = True

    def resume_writing(self):
        assert self._paused
        self._paused = False

        waiter = self._drain_waiter
        if waiter is not None:
            self._drain_waiter = None
            if not waiter.done():
                waiter.set_result(None)

    def connection_lost(self, exc):
        self._connection_lost = True
        # Wake up the writer if currently paused.
        if not self._paused:
            return
        waiter = self._drain_waiter
        if waiter is None:
            return
        self._drain_waiter = None
        if waiter.done():
            return
        if exc is None:
            waiter.set_result(None)
        else:
            waiter.set_exception(exc)

    async def drain(self):
        if self._connection_lost:
            raise ConnectionResetError('Connection lost')
        if not self._paused:
            return
        waiter = self._drain_waiter
        assert waiter is None or waiter.cancelled()
        waiter = self.loop.create_future()
        self._drain_waiter = waiter
        await waiter

    async def write(self, data: bytes, *, drain: bool = False) -> None:
        extra = {
            'protocol': self,
            'data': data
        }
        if self.state is BaseProtocolState.CLOSED:
            raise ConnectionClosedError(
                'Attempt to write to a closed/closing transport',
                extra)

        self.transport.write(data)
        if drain:
            await self.drain()

    def closing_connection(self, exc) -> None:
        pass


def taskify(func):
    @functools.wraps(func)
    def callback(protocol, *args, **kwargs):
        protocol.loop.create_task(func(protocol, *args, **kwargs))
    return callback



from .client import WebSocketClient
from .exceptions import (BrokenHandshakeError, ParserInvalidDataError,
                         WsaioError)
from .http import (Headers, HTTPRequest, HTTPRequestProtocol,
                   HTTPResponse, HTTPResponseProtocol)
from .protocol import BaseProtocol, taskify
from .websocket import (WebSocketCloseCode, WebSocketFrame, WebSocketOpcode,
                        WebSocketProtocol, WebSocketState)



from __future__ import annotations

import enum
import os
import struct
from typing import ByteString

from .exceptions import ParserInvalidDataError
from .protocol import BaseProtocol
from .utils import ensure_length


class WebSocketOpcode(enum.IntEnum):
    CONTINUATION = 0x0
    TEXT = 0x1
    BINARY = 0x2
    CLOSE = 0x8
    PING = 0x9
    PONG = 0xA


CONTROL_OPCODES = (WebSocketOpcode.CLOSE, WebSocketOpcode.PING,
                   WebSocketOpcode.PONG)


class WebSocketCloseCode(enum.IntEnum):
    # 0 - 999 NOT USED
    NORMAL_CLOSURE = 1000  # Server <-> Client
    GOING_AWAY = 1001  # Server <-> Client
    PROTOCOL_ERROR = 1002  # Server <-> Client
    UNSUPPORTED_DATA = 1003  # Server <-> Client
    NO_STATUS_RECEIVED = 1005  # Server ~ Client (Reserved for "applications")
    ABNORMAL_CLOSURE = 1006  # Server ~ Client (Reserved for "applications")
    INVALID_PAYLOAD_DATA = 1007  # Server <-> Client
    POLICY_VIOLATION = 1008  # Server <-> Client
    MESSAGE_TOO_BIG = 1009  # Server <-> Client
    MANDATORY_EXTENSION = 1010  # Server <-(>?) Client
    INTERNAL_SERVER_ERROR = 1011  # Server -> Client
    TLS_HANDSHAKE = 1015  # Server ~ Client (Reserved for "applications")
    # XXX: What's the difference between an application and an endpoint?
    # 1000 - 2999 RESERVED FOR WEBSOCKET PROTOCOL
    # 3000 - 3999 RESERVED FOR LIBRARIES, FRAMEWORKS AND APPS
    # 4000 - 4999 RESERVED FOR PRIVATE USE


class WebSocketState(enum.IntEnum):
    HANDSHAKING = 4


class WebSocketProtocol:
    def __init__(self):
        self.extensions = []

    def strstate(self):
        strstate = BaseProtocol.strstate(self)
        if strstate is not None:
            return strstate
        elif self.state is WebSocketState.HANDSHAKING:
            return 'handshaking'

    def ws_connected(self) -> None:
        pass

    def ws_frame_received(self, frame: WebSocketFrame) -> None:
        pass

    def ws_binary_received(self, data: bytes) -> None:
        pass

    def ws_text_received(self, data: str) -> None:
        pass

    def ws_ping_received(self, data: bytes) -> None:
        pass

    def ws_pong_received(self, data: bytes) -> None:
        pass

    def ws_close_received(self, code: int, data: bytes) -> None:
        pass


class WebSocketFrame:
    SHORT_LENGTH = struct.Struct('!H')
    LONGLONG_LENGTH = struct.Struct('!Q')

    def __init__(self, *, opcode: WebSocketOpcode, fin: bool = True,
                 rsv1: bool = False, rsv2: bool = False, rsv3: bool = False,
                 data: ByteString):
        self.opcode = opcode
        self.fin = fin
        self.rsv1 = rsv1
        self.rsv2 = rsv2
        self.rsv3 = rsv3
        self.data = data

    def __repr__(self) -> str:
        attrs = ('fin', 'rsv1', 'rsv2', 'rsv3', 'opcode')
        s = ', '.join(f'{name}={getattr(self, name)!r}' for name in attrs)
        return f'<{self.__class__.__name__} {s}>'

    @staticmethod
    def mask(data: ByteString, mask: bytes) -> bytearray:
        data = bytearray(data)
        for i in range(len(data)):
            data[i] ^= mask[i % 4]
        return data

    def encode(self, masked: bool = False) -> bytearray:
        buffer = bytearray(2)
        buffer[0] = ((self.fin << 7)
                     | (self.rsv1 << 6)
                     | (self.rsv2 << 5)
                     | (self.rsv3 << 4)
                     | self.opcode)
        buffer[1] = masked << 7

        length = len(self.data)
        if length < 126:
            buffer[1] |= length
        elif length < 2 ** 16:
            buffer[1] |= 126
            buffer.extend(self.SHORT_LENGTH.pack(length))
        else:
            buffer[1] |= 127
            buffer.extend(self.LONGLONG_LENGTH.pack(length))

        if masked:
            mask = os.urandom(4)
            buffer.extend(mask)
            data = self.mask(self.data, mask)
        else:
            data = self.data

        buffer.extend(data)

        return buffer

    @classmethod
    def parser(cls, protocol: WebSocketProtocol):
        data = yield
        position = 0
        fragmented_frame = None
        fragment_buffer = bytearray()

        while True:
            data = data[position:]
            position = 0

            data = yield from ensure_length(data, position + 1)
            fbyte = data[position]
            position += 1

            data = yield from ensure_length(data, position + 1)
            sbyte = data[position]
            position += 1

            masked = (sbyte >> 7) & 1
            length = sbyte & ~(1 << 7)

            if length > 125:
                if length == 126:
                    strct = cls.SHORT_LENGTH
                elif length == 127:
                    strct = cls.LONGLONG_LENGTH

                data = yield from ensure_length(data, position + strct.size)
                length, = strct.unpack_from(data, position)
                position += strct.size

            if masked:
                data = yield from ensure_length(data, position + 4)
                mask = data[position:position + 4]
                position += 4

            data = yield from ensure_length(data, position + length)
            payload = data[position:position + length]
            position += length

            if masked:
                payload = cls.mask(data, mask)

            frame = cls(opcode=WebSocketOpcode(fbyte & 0xF),
                        fin=(fbyte >> 7) & 1, rsv1=(fbyte >> 6) & 1,
                        rsv2=(fbyte >> 5) & 1, rsv3=(fbyte >> 4) & 1,
                        data=payload)
            protocol.ws_frame_received(frame)

            extra = {
                'frame': frame,
                'protocol': protocol
            }

            if not protocol.extensions:
                if any((frame.rsv1, frame.rsv2, frame.rsv3)):
                    extra['close_code'] = WebSocketCloseCode.PROTOCOL_ERROR
                    raise ParserInvalidDataError(
                        'Received rsv1, rsv2 or rsv3 but no extensions '
                        'were negotiated', extra)

            if frame.opcode in CONTROL_OPCODES:
                if len(frame.data) > 125:
                    extra['close_code'] = WebSocketCloseCode.PROTOCOL_ERROR
                    raise ParserInvalidDataError(
                        'Received control frame with payload length > 125 '
                        f'({len(frame.data)})', extra)

                elif not frame.fin:
                    extra['close_code'] = WebSocketCloseCode.PROTOCOL_ERROR
                    raise ParserInvalidDataError(
                        'Received fragmented control frame', extra)

                elif frame.opcode is WebSocketOpcode.PING:
                    protocol.ws_ping_received(frame.data)

                elif frame.opcode is WebSocketOpcode.PONG:
                    protocol.ws_pong_received(frame.data)

                elif frame.opcode is WebSocketOpcode.CLOSE:
                    close_clode = int.from_bytes(frame.data[:2], 'big')
                    data = frame.data[2:]
                    protocol.ws_close_received(close_clode, data)
            else:
                if fragmented_frame is not None:
                    if frame.opcode is not WebSocketOpcode.CONTINUATION:
                        extra['close_code'] = WebSocketCloseCode.PROTOCOL_ERROR
                        raise ParserInvalidDataError(
                            f'Expected opcode {WebSocketOpcode.CONTINUATION} '
                            f'(got {frame.opcode})',
                            extra
                        )
                    else:
                        fragment_buffer.extend(frame.data)
                        if frame.fin:
                            frame = fragmented_frame
                            fragmented_frame = None
                            frame.data = bytes(fragment_buffer)

                elif not frame.fin:
                    fragment_buffer.clear()
                    fragment_buffer.extend(frame.data)
                    fragmented_frame = frame

                if frame.opcode is WebSocketOpcode.TEXT:
                    try:
                        string = frame.data.decode()
                    except UnicodeDecodeError as e:
                        extra['close_code'] = (
                            WebSocketCloseCode.INVALID_PAYLOAD_DATA
                        )
                        raise ParserInvalidDataError(str(e), extra)
                    protocol.ws_text_received(string)

                elif frame.opcode is WebSocketOpcode.BINARY:
                    protocol.ws_binary_received(frame.data)



class WsaioError(Exception):
    def __init__(self, msg, extra):
        super().__init__(msg)
        self._extra = extra

    def get_extra(self, key, default=None):
        return self._extra.get(key, default)


class BrokenHandshakeError(WsaioError):
    pass


class ParserInvalidDataError(WsaioError):
    pass


class ConnectionClosedError(WsaioError, ConnectionError):
    pass



import setuptools

with open('README.md', 'r') as fp:
    readme = fp.read()

setuptools.setup(
    name='wsaio',
    version='0.0.3',
    author='asleep-cult',
    description='An event-driven WebSocket library for Python.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/asleep-cult/wsaio',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.7",
)
